#include <Parser.H>
#include <string>
#include <stdio.h>
#include <fstream>
#include <Vector>


namespace GRF
{

Parser::Parser
(
   const char * _fileName,
   bool         _verbose
)
:  fileName(_fileName)
,  verbose(_verbose)
,  stack(NULL)
,  nLights(0L)
,  nVertexs(0L)
,  nFaces(0L)
{

   OpenStream(fileName);

   // get look-ahead token for the first time

   LoadNextToken() ;

   // a grf file is a (maybe empty) list of blocks:

   while ( tk.Type() != tkEndOfInput )
      ReadBlock() ;

   if ( verbose )
      std::cout << std::endl << "end parsing file (" << fileName << ")"
           << std::endl << "total vertexs    = " << nVertexs
           << std::endl << "total faces      = " << nFaces
           << std::flush ;
}

void Parser::OpenStream
(
   const std::string & fileName
)
{
   std::cout << std::endl << "opening include file (" << fileName << ") ..." << std::flush ;

   InputStack * res ;

   res = new InputStack ;

   if ( res == NULL )
        throw ParserError("new failed","Parser::OpenStream");

   res->is = new std::ifstream(fileName.c_str()) ;

   if ( res->is == NULL )
      throw ParserError("cannot alloc or initialize","creating parser") ;

   if ( ! res->is->good() )
   {
      std::string s = "cannot open file (" + fileName + ") ";
      throw ParserError(s,"Parser::OpenStream") ;
   }


   res->next = stack ;
   stack = res ;
}

void Parser::CloseStream()
{
   InputStack * tmp = stack ;

   stack = stack->next ;

   delete tmp ;
}

void Parser::LoadNextToken()
{
  if ( stack == NULL )
     throw ParserError("internal error: empty stack","Parser::LoadNextToken");

  *(stack->is) >> tk ;

  while ( tk.Type() == tkEndOfInput && stack->next != NULL )
  {
      CloseStream() ;
      *(stack->is) >> tk ;
  }
}

//************************************************************

void Parser::ReadBlock()
{
   // a block is either a mesh or a light
   //

   if ( tk.Type() == tkBegin )
       ReadMesh() ;
   else if ( tk.Type() == tkLight )
       ReadLight() ;
   else if ( tk.Type() == tkInclude )
       ReadInclude() ;
   else if ( tk.Type() == tkTransform )
       ReadTransformBlock() ;
   else
       throw ParserError("'begin mesh' or 'light begin' expected","ReadBlock") ;
}

//************************************************************

void Parser::ReadTransformBlock()
{
   ReadSingleToken(tkTransform) ;

   while ( tk.Type() == tkScale ||
           tk.Type() == tkRotation ||
           tk.Type() == tkTranslation
         )
   {

   }
}

//************************************************************

void Parser::ReadInclude()
{
   ReadSingleToken(tkInclude);

   if ( tk.Type() != tkLiteral )
      throw ParserError("literal expected after 'include' (did you used '\"' ?)","Parser::ReadInclude");

   std::string fileName = tk.ValueString() ;

   OpenStream(fileName);
   ReadSingleToken(tkLiteral);

}

//************************************************************

void Parser::ReadMesh()
{

   pCurrentFaceColorRGB = NULL ;

   meshVertexIndexBase = nVertexs ;
   nMeshVertexs = 0L ;

   ReadSingleToken( tkBegin ) ;
   ReadSingleTokenOpt( tkMesh ) ;

   ReadMeshVertexs() ;
   ReadMeshFaces() ;

   ReadSingleToken( tkEnd ) ;
   ReadSingleTokenOpt( tkMesh ) ;
}

//************************************************************

void Parser::ReadLight()
{
   ReadSingleToken(tkLight);
   ReadSingleToken(tkBegin) ;

   double org[3], edge0[3], edge1[3], rgb[3] ;

   ReadVector3(org);
   ReadVector3(edge0);
   ReadVector3(edge1);

   rgb[0] = 1.0 ;
   rgb[1] = 1.0 ;
   rgb[2] = 1.0 ;

   if ( tk.Type() == tkRGB )
   {
      ReadSingleToken(tkRGB) ;
      ReadVector3(rgb) ;
   }

   ReadSingleToken(tkEnd);
   //ReadSingleTokenOpt(tkLight);

   Light l(org,edge0,edge1,rgb) ;

   lights.push_back(l) ;

   nLights++ ;

}

//************************************************************

void Parser::ReadSingleToken
(
   TokenType type
)
{
   if ( tk.Type() != type )
      throw ParserError("expected token not found","Parser::ReadSingleToken") ;

   LoadNextToken() ;
}

//************************************************************

void Parser::ReadSingleNumericToken()
{
   if ( tk.Type() != tkNumericInteger &&
        tk.Type() != tkNumericDouble )
      throw ParserError("expected numeric token not found","Parser::ReadSingleNumericToken") ;

   LoadNextToken() ;
}

//************************************************************

void Parser::ReadSingleTokenOpt
(
   TokenType type
)
{
   if ( tk.Type() == type )
      LoadNextToken() ;
}

//************************************************************

void Parser::ReadMeshVertexs()
{
   ReadSingleToken( tkBegin ) ;
   ReadSingleTokenOpt( tkVertexs ) ;

   while( tk.Type() == tkOpenPar ||
          tk.Type() == tkNormal  ||
          tk.Type() == tkIrad )
   {
      ReadVertex() ;
   }


   ReadSingleToken( tkEnd ) ;
   ReadSingleTokenOpt( tkVertexs ) ;
}

//************************************************************

void Parser::ReadMeshFaces()
{
   ReadSingleToken( tkBegin ) ;
   ReadSingleTokenOpt( tkFaces ) ;

   while( 1 )
   {
      if ( tk.Type() == tkNumericInteger )
         ReadFace() ;
      else if ( tk.Type() == tkOpenPar )
         ReadFace2() ;
      else if ( tk.Type() == tkRGB )
         ReadFacesColor() ;
      else
         break ;
   }

   ReadSingleToken( tkEnd ) ;
   ReadSingleTokenOpt( tkFaces ) ;
}

//************************************************************

void Parser::ReadFacesColor()
{
   ReadSingleToken(tkRGB) ;

   ReadVector3(currentFaceColorRGB) ;

   /**
   std::cout << std::endl << "** change face rgb: ("
       << currentFaceColorRGB[0] << ","
       << currentFaceColorRGB[1] << ","
       << currentFaceColorRGB[2] << ")" ;
   **/

   pCurrentFaceColorRGB = currentFaceColorRGB ;
}

//************************************************************

void Parser::ReadVertex()
{
   double vp[3], vn[3], irad[3] ;

   Vertex v ;

   if ( tk.Type() == tkNormal )
   {
      ReadSingleToken(tkNormal) ;
      ReadVector3(vn) ;
      v.SetNormal(vn) ;
   }

   if ( tk.Type() == tkIrad )
   {
      ReadSingleToken(tkIrad);
      ReadVector3(irad);
      v.SetIradiance(irad);
   }

   ReadVector3(vp) ;
   v.SetPosition(vp) ;

   vertexs.push_back( v ) ;

   nVertexs++ ;
   nMeshVertexs ++ ;
}

//************************************************************

void Parser::ReadVector3
(
   double v[3]
)
{
   ReadSingleToken(tkOpenPar) ;

   v[0] = tk.ValueDouble() ;
   ReadSingleNumericToken();
   ReadSingleToken(tkComma) ;

   v[1] = tk.ValueDouble() ;
   ReadSingleNumericToken();
   ReadSingleToken(tkComma) ;

   v[2] = tk.ValueDouble() ;
   ReadSingleNumericToken();
   ReadSingleToken(tkClosePar);
}

//************************************************************

void Parser::ReadFace()
{
   long nv = tk.ValueInteger() ;

   ReadSingleToken(tkNumericInteger) ;

   if ( nv < 0 || nv > 10000 )
      throw ParserError("invalid number of vertexs in face","ReadFace");

   std::vector<unsigned long> fv(nv) ;

   for( int i = 0 ; i < nv ; i++ )
   {
      long iv = tk.ValueInteger() ;

      if ( iv < 0L || nMeshVertexs <= (unsigned long)iv )
      {
         char s[200] ;
         sprintf(s,"vertex index out of bound (index=%ld,num.mesh.vertexs=%ld)",iv,nMeshVertexs) ;

         throw ParserError(s,"raised while reading vertex list in a face (Parser::ReadFace)") ;
      }

      fv[i] = (unsigned long)iv + meshVertexIndexBase ;

      ReadSingleToken(tkNumericInteger) ;
   }

   faces.push_back( Face(fv,pCurrentFaceColorRGB) ) ;
   nFaces ++ ;
}

//************************************************************

void Parser::ReadFace2()
{


   ReadSingleToken(tkOpenPar) ;

   std::vector<unsigned long> fv ;

   while( tk.Type() != tkClosePar )
   {
      long iv = tk.ValueInteger() ;

      ReadSingleToken(tkNumericInteger) ;

      if ( iv < 0L || nMeshVertexs <= (unsigned long)iv )
      {
         char s[200] ;
         sprintf(s,"vertex index out of bound ( index = %ld , num.mesh.vertexs = %ld )",iv,nMeshVertexs) ;

         throw ParserError(s,"raised while reading vertex list in a face (Parser::ReadFace2)") ;
      }

      fv.push_back(  (unsigned long)iv + meshVertexIndexBase ) ;

      if ( tk.Type() == tkComma )
         ReadSingleToken(tkComma) ;
      else
      if ( tk.Type() == tkClosePar )
         break ;
      else
      if ( tk.Type() != tkNumericInteger )
         throw ParserError("integer index expected in face","Parser::ReadFace2");
   }

   ReadSingleToken( tkClosePar ) ;

   faces.push_back( Face(fv,pCurrentFaceColorRGB) ) ;
   nFaces ++ ;
}

//************************************************************

Parser::~Parser()
{


}

//************************************************************

const Face * Parser::GetFace
(
   unsigned long iFace
)
   const
{
   if ( iFace >=  nFaces )
   {
      char s[200] ;
      sprintf(s,"face index out of bounds (index=%ld,num.faces=%ld)",iFace,nFaces) ;

      throw ParserError(s,"Parser::GetFace") ;
   }

   return & ( faces[iFace] ) ;
}

//************************************************************

const Vertex * Parser::GetVertex
(
   unsigned long iVertex
)
   const
{

   if ( iVertex >=  nVertexs )
   {
     char s[200] ;
     sprintf(s,"vertex index out of bounds (index=%ld,num.vertexs=%ld)",iVertex,nVertexs) ;

     throw ParserError(s,"Parser::GetVertex") ;
   }

   return & ( vertexs[iVertex] ) ;
}

//*************************************************************

unsigned long Parser::NVertexs
()
   const
{
   return nVertexs ;
}

//*****************************************************************

unsigned long Parser::NFaces
()
   const
{
   return nFaces ;
}

//*************************************************************

const Light * Parser::GetLight
(
   unsigned long iLight
)
   const
{
   if ( iLight >=  nLights )
   {
     char s[200] ;
     sprintf(s,"light index out of bounds (index=%ld,num.lights=%ld)",
                  iLight,nLights) ;

     throw ParserError(s,"Parser::GetLight") ;
   }

   return &( lights[iLight] ) ;
}

unsigned long Parser::NLights
()
   const
{
   return nLights ;
}


} ; // namespace GRF
